# Generated by Django 2.0.4 on 2018-04-26 18:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("djnexmo", "0002_auto_20180426_1521")]

    operations = [
        migrations.AlterField(
            model_name="smsmessagepart",
            name="message_id",
            field=models.CharField(max_length=32, unique=True),
        ),
        migrations.AlterUniqueTogether(
            name="smsmessagepart", unique_together={("concat_ref", "concat_part")}
        ),
    ]




from datetime import datetime, timezone
import json

import django
from django.db import transaction

import djnexmo.decorators as d
import djnexmo.models as models

from random import shuffle
from unittest.mock import MagicMock, call, sentinel

import pytest


@pytest.fixture(name="partial_message")
def partial_message_fixture():
    return {
        "concat": "true",
        "concat-part": "1",
        "concat-ref": "78",
        "concat-total": "9",
        "keyword": "LOREM",
        "message-timestamp": "2018-04-24 14:05:19",
        "messageId": "0B000000D0EBB58D",
        "msisdn": "447700900419",
        "nonce": "0a455d75-459e-446c-9072-698728516d7c",
        "sig": "c4bc6301949691b6093772ba246a35eb",
        "text": "Lorem Ipsum is simply dummy text of the printing and typesetting in",
        "timestamp": "1524578719",
        "to": "447700900996",
        "type": "unicode",
    }


@pytest.fixture(name="complete_message")
def complete_message_fixture():
    return {
        "keyword": "THIS",
        "message-timestamp": "2018-04-24 14:05:19",
        "messageId": "0B000000D0EBB58D",
        "msisdn": "447700900419",
        "nonce": "0a455d75-459e-446c-9072-698728516d7c",
        "sig": "266c3734d3cf9baf6f293dae148e14f5",
        "text": "This is complete!",
        "timestamp": "1524578719",
        "to": "447700900996",
        "type": "unicode",
    }


def test_parse_part(partial_message):
    """ Ensure we can parse an incoming json sms request. """
    sms = d.IncomingSMSSchema().load(partial_message)
    assert isinstance(sms, d.IncomingSMS)
    assert sms.concat == True
    assert sms.concat_part == 1
    assert sms.concat_total == 9
    assert sms.concat_ref == "78"
    assert sms.keyword == "LOREM"
    assert sms.message_timestamp == datetime(
        2018, 4, 24, 14, 5, 19, tzinfo=timezone.utc
    )
    assert sms.message_id == "0B000000D0EBB58D"
    assert sms.msisdn == "447700900419"
    assert sms.to == "447700900996"
    assert (
        sms.text
        == "Lorem Ipsum is simply dummy text of the printing and typesetting in"
    )
    assert sms.timestamp == datetime(2018, 4, 24, 14, 5, 19, tzinfo=timezone.utc)
    assert sms.type == "unicode"


def test_to_model(partial_message):
    """ Ensure translating from the parsed object to the django model works. """
    sms = d.IncomingSMSSchema().load(partial_message).to_model()
    assert isinstance(sms, models.SMSMessagePart)
    assert not hasattr(sms, "concat")  # Just make sure we're looking at the model.
    assert sms.concat_part == 1
    assert sms.concat_total == 9
    assert sms.concat_ref == "78"
    assert sms.keyword == "LOREM"
    assert sms.message_timestamp == datetime(
        2018, 4, 24, 14, 5, 19, tzinfo=timezone.utc
    )
    assert sms.message_id == "0B000000D0EBB58D"
    assert sms.msisdn == "447700900419"
    assert sms.to == "447700900996"
    assert (
        sms.text
        == "Lorem Ipsum is simply dummy text of the printing and typesetting in"
    )
    assert sms.timestamp == datetime(2018, 4, 24, 14, 5, 19, tzinfo=timezone.utc)
    assert sms.type == "unicode"


@pytest.mark.django_db
def test_duplicate_partials(partial_message):
    """ Ensure saving the same message part twice fails. """
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "There should be no parts stored before the test."
    sms = d.IncomingSMSSchema().load(partial_message).to_model()
    sms.save()

    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 1
    ), "Part should be saved."

    with pytest.raises(django.db.IntegrityError), transaction.atomic():
        sms2 = d.IncomingSMSSchema().load(partial_message).to_model()
        sms2.save()

    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 1
    ), "Duplicate part should be not be saved."


@pytest.mark.django_db
def test_decorator_partials(rf, partial_message):
    """ Ensure partial sms messages are handled corrrectly. """
    request = rf.post(
        "/sms/incoming",
        content_type="application/json",
        data=json.dumps(partial_message),
    )
    view = MagicMock()
    decorated = d.sms_webhook()(view)

    # Before doing anything, make sure we don't already have any parts saved:
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "There should be no parts stored before the test."

    response = decorated(request)

    assert response.status_code == 200, "Request should be successful"
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 1
    ), "Part should be saved."

    # Ensure a duplicate part doesn't also get saved:
    response = decorated(request)

    assert response.status_code == 200, "Request should be successful"
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 1
    ), "No duplicate part should be saved."
    assert view.mock_calls == [], "Wrapped view should not be called."


@pytest.mark.django_db
def test_decorator_complete(rf, complete_message):
    request = rf.post(
        "/sms/incoming",
        content_type="application/json",
        data=json.dumps(complete_message),
    )

    view = MagicMock(return_value=sentinel.response)

    decorated = d.sms_webhook()(view)
    response = decorated(request)
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "Message should not be saved."
    assert view.mock_calls == [call(request)], "Underlying view should be called."
    assert (
        response == sentinel.response
    ), "Result should be response from the underlying view."


@pytest.mark.django_db
def test_decorator_complete_bad_sig(rf, complete_message):
    complete_message["sig"] = "not-valid"
    request = rf.post(
        "/sms/incoming",
        content_type="application/json",
        data=json.dumps(complete_message),
    )
    view = MagicMock(return_value=sentinel.response)

    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "There should be no parts stored before the test."

    validating = d.sms_webhook(validate_signature=True)(view)
    response = validating(request)

    assert view.mock_calls == [], "Wrapped view should not be called."
    assert (
        response.status_code == 403
    ), "Decorator should return 403, permission denied."
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "No part should be saved."

    nonvalidating = d.sms_webhook(validate_signature=False)(view)
    response = nonvalidating(request)

    assert (
        response == sentinel.response
    ), "Result should be response from the underlying view."
    assert view.mock_calls == [call(request)], "Underlying view should be called."
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "No part should be saved."


@pytest.mark.django_db
def test_decorator_partial__bad_sig(rf, partial_message):
    partial_message["sig"] = "not-valid"
    request = rf.post(
        "/sms/incoming",
        content_type="application/json",
        data=json.dumps(partial_message),
    )

    view = MagicMock(return_value=sentinel.response)

    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "There should be no parts stored before the test."

    validating = d.sms_webhook(validate_signature=True)(view)
    response = validating(request)

    assert view.mock_calls == [], "Wrapped view should not be called."
    assert (
        response.status_code == 403
    ), "Decorator should return 403, permission denied."
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 0
    ), "No part should be saved."

    nonvalidating = d.sms_webhook(validate_signature=False)(view)
    response = nonvalidating(request)

    assert (
        response.status_code == 200
    ), "OK response should be returned by the decorator."
    assert view.mock_calls == [], "Wrapped view should not be called."
    assert (
        len(models.SMSMessagePart.objects.filter(concat_ref="78")) == 1
    ), "Part should be saved."


@pytest.mark.django_db
def test_decorator_multipart(rf):
    parts = ["This", " is", " a ", "multipart", " message"]

    view = MagicMock(return_value=sentinel.response)
    webhook = d.sms_webhook(validate_signature=False)(view)

    requests = [
        rf.post(
            "/sms/incoming",
            content_type="application/json",
            data=json.dumps(
                {
                    "concat": "true",
                    "concat-part": "{index_1b}".format(index_1b=index + 1),
                    "concat-ref": "78",
                    "concat-total": "{parts_len}".format(parts_len=len(parts)),
                    "keyword": "LOREM",
                    "message-timestamp": "2018-04-24 14:05:19",
                    "messageId": "0B000000D0EBB58{index}".format(index=index),
                    "msisdn": "447700900419",
                    "text": part,
                    "timestamp": "1524578719",
                    "to": "447700900996",
                    "type": "unicode",
                }
            ),
        )
        for index, part in enumerate(parts)
    ]
    shuffle(requests)

    for request in requests[:-1]:
        response = webhook(request)
        assert response.status_code == 200

    response = webhook(requests[-1])
    assert response is sentinel.response

    assert len(view.mock_calls) == 1
    name, args, kwargs = view.mock_calls[0]
    assert args[0].sms.text == "This is a multipart message"



from invoke import task


@task
def build(c):
    # c.run("pandoc -f markdown -t rst -o README.rst README.md")
    c.run("python setup.py bdist_wheel sdist")


@task
def clean(c):
    c.run("rm -rf README.rst build dist")



# Generated by Django 2.0.4 on 2018-04-26 10:54

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SMSMessagePart",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("message_id", models.CharField(max_length=32)),
                ("msisdn", models.CharField(max_length=24)),
                ("to", models.CharField(max_length=24)),
                ("text", models.CharField(max_length=160)),
                ("data", models.BinaryField(max_length=160)),
                ("udh", models.BinaryField(max_length=160)),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("text", "Text"),
                            ("unicode", "Unicode"),
                            ("binary", "Binary"),
                        ],
                        max_length=7,
                    ),
                ),
                ("keyword", models.CharField(max_length=160)),
                ("message_timestamp", models.DateTimeField()),
                ("timestamp", models.DateTimeField()),
                ("concat_part", models.IntegerField()),
                ("concat_ref", models.CharField(max_length=32)),
                ("concat_total", models.IntegerField()),
            ],
        )
    ]



"""
djnexmo.decorators - decorators that make it easier to work with Nexmo.


"""

from datetime import datetime, timezone
from functools import wraps
import json

from django.db import IntegrityError, transaction
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

import attr
from marshmallow import Schema, fields, post_load
import pytz

from .models import SMSMessagePart

from . import client


TZ_LONDON = pytz.timezone('Europe/London')

@attr.s
class IncomingSMS:
    """ Object representing an incoming SMS message or part-message, parsed from JSON. """
    message_id = attr.ib(type=str)
    msisdn = attr.ib(type=str)
    to = attr.ib(type=str)
    type = attr.ib(
        type=str, validator=attr.validators.in_(["text", "unicode", "binary"])
    )
    message_timestamp = attr.ib(type=datetime)
    timestamp = attr.ib(type=datetime)

    keyword = attr.ib(type=str, default=None)
    text = attr.ib(type=str, default=None)
    # TODO: Need to do something with these:
    data = attr.ib(default=None)
    udh = attr.ib(default=None)

    concat = attr.ib(type=bool, default=False)
    concat_part = attr.ib(type=int, default=None)
    concat_ref = attr.ib(type=str, default=None)
    concat_total = attr.ib(type=int, default=None)

    def reply(self, text, type="text"):
        client.send_message(
            {"to": self.msisdn, "from": self.to, "text": text, "type": type}
        )

    def to_model(self):
        data = {a.name: getattr(self, a.name) for a in attr.fields(self.__class__)}
        del data["concat"]
        return SMSMessagePart(**data)


class Timestamp(fields.Field):
    """ Marshmallow Field to convert a UTC unix time into a UTC timezone-aware datetime. """

    def _deserialize(self, value, attr, data):
        # `value` is a UNIX timestamp in Europe/London tz, as a string 🙄
        return datetime.utcfromtimestamp(int(value)).replace(tzinfo=timezone.utc)


class IncomingSMSSchema(Schema):
    """ Marshmallow schema to map from Nexmo incoming SMS JSON to an IncomingSMS instance. """
    msisdn = fields.Str()
    to = fields.Str()
    message_id = fields.Str(data_key="messageId")
    text = fields.Str()
    type = fields.Str(data_key="type")
    keyword = fields.Str()
    message_timestamp = fields.DateTime(
        data_key="message-timestamp", format="%Y-%m-%d %H:%M:%S"
    )
    timestamp = Timestamp()

    concat = fields.Bool()
    concat_part = fields.Int(data_key="concat-part")
    concat_ref = fields.Str(data_key="concat-ref")
    concat_total = fields.Int(data_key="concat-total")

    # TODO: Need to do something with these:
    data = fields.Str()
    udh = fields.Str()

    @post_load
    def make_sms(self, data):
        d = data["message_timestamp"]
        if d.tzinfo is None:
            data["message_timestamp"] = d.replace(tzinfo=timezone.utc)
        return IncomingSMS(**data)


incoming_sms_parser = IncomingSMSSchema()


def sms_webhook(func=None, *, validate_signature=True):
    """
    A decorator for views which respond to incoming SMS messages.

    Example::

        @sms_webhook()
        def birthday_rsvp_view(request):
            # Access the `IncomingSMS` object via `request.sms`
            if request.sms.text.upper() == 'YES':
                BirthdayResponses(sender_tel=request.sms.msisdn, attending=True).save()

    Behind the scenes, a couple of things are done for you:

    * The signature is verified against your signature secret, defined in
      `settings.NEXMO_SIGNATURE_SECRET`. If you don't want the signature to be
      verified, call with `sms_webhook` with `validate_signature=False`
    * Messages sent as multiple parts are stored in the database until all
      parts are available. The underlying view is only called once all parts
      are available and have been merged into a single `IncomingSMS` instance.
    """

    def decorator(func):

        @wraps(func)
        @csrf_exempt
        @require_POST
        def inner(request, *args, **kwargs):
            if request.content_type != "application/json":
                return HttpResponse("Unsupported request content-type.", status=415)
            try:
                data = json.loads(request.body.decode("utf-8"))
            except json.JSONDecodeError:
                return HttpResponse("Invalid JSON payload provided.", status=400)
            if not validate_signature or client.check_signature(data):
                if data.get("concat") == "true":
                    return _handle_message_part(request, data, func, args, kwargs)
                else:
                    request.sms = incoming_sms_parser.load(data)
                    return func(request, *args, **kwargs)
            else:
                return HttpResponse(
                    "Invalid signature.", status=403, reason="Invalid signature."
                )

        return inner

    if func is not None:
        return decorator(func)
    else:
        return decorator


def _handle_message_part(request, data, wrapped_func, args, kwargs):
    # Put it in the database.
    try:
        with transaction.atomic():
            incoming_sms = incoming_sms_parser.load(data)
            incoming_sms.to_model().save()
    except IntegrityError:
        return HttpResponse("Partial message already stored.")

    # Then query if we have all the parts
    matching_parts = SMSMessagePart.objects.filter(concat_ref=incoming_sms.concat_ref)
    count = matching_parts.count()
    expected = incoming_sms.concat_total
    if count == expected:
        # If we have all the parts then create a FrankenSMS from the pieces,
        # and call the wrapped view function:
        text = "".join(part.text for part in matching_parts)
        request.sms = IncomingSMS(
            msisdn=incoming_sms.msisdn,
            to=incoming_sms.to,
            message_id=incoming_sms.message_id,
            text=text,
            type=incoming_sms.type,
            keyword=incoming_sms.keyword,
            message_timestamp=incoming_sms.message_timestamp,
            timestamp=incoming_sms.timestamp,
            concat=False,
            concat_ref=incoming_sms.concat_ref,
        )
        matching_parts.delete()
        return wrapped_func(request, *args, **kwargs)
    else:
        return HttpResponse("Partial message received.")



from django.db import models


class SMSMessagePart(models.Model):

    class Meta:
        ordering = ("-concat_ref", "concat_part")
        unique_together = ("concat_ref", "concat_part")
        verbose_name = "Message Part"
        verbose_name_plural = "Message Parts"

    concat_ref = models.CharField(max_length=32, db_index=True)

    message_id = models.CharField(max_length=32, unique=True)
    msisdn = models.CharField(max_length=24)
    to = models.CharField(max_length=24)

    text = models.CharField(max_length=160, null=True)
    # TODO: Need to do something with these:
    data = models.BinaryField(max_length=160, null=True)
    udh = models.BinaryField(max_length=160, null=True)

    type = models.CharField(
        max_length=7,
        choices=[("text", "Text"), ("unicode", "Unicode"), ("binary", "Binary")],
    )
    keyword = models.CharField(max_length=160, null=True)
    message_timestamp = models.DateTimeField()
    timestamp = models.DateTimeField()

    concat_part = models.IntegerField()
    concat_ref = models.CharField(max_length=32)
    concat_total = models.IntegerField()

    def __str__(self):
        return (
            "Message {self.concat_ref}: {self.text!r} (Part {self.concat_part} of {self.concat_total})".format(
                self=self,
            )
        )



from django.contrib import admin

from .models import SMSMessagePart


@admin.register(SMSMessagePart)
class SMSMessagePartAdmin(admin.ModelAdmin):
    list_display = ("msisdn", "to", "__str__")
    list_display_links = ("__str__",)
    search_fields = ("msisdn",)
    view_on_site = False



# Generated by Django 2.0.4 on 2018-04-30 15:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("djnexmo", "0004_auto_20180427_1517")]

    operations = [
        migrations.AlterModelOptions(
            name="smsmessagepart",
            options={
                "ordering": ("-concat_ref", "concat_part"),
                "verbose_name": "Message Part",
                "verbose_name_plural": "Message Parts",
            },
        )
    ]
